#!/usr/bin/python

"""
# bayesian inversion
#
# Solve the equation
#  shat = sprior + HQT*inverse(HQHT+R)*(z-hsp)
#
# This script calculates the HQ and HQHT parts of the equation.
#
# This version uses python multiprocessing to split up the jobs.
# Creates a hqht block for each process, than combines them to
# get one final hqht
"""

from __future__ import print_function

import os
import sys
import datetime
import multiprocessing
import optparse
import numpy

from scipy.sparse import csr_matrix

import utils

####################################################################
def makeHQBlocks(config, start, end, sigmafile, TC=None, SC=None):
	"""
	Compute transpose(HQH).
	Use the method of Yadav and Michalak, 2012 for expressing Q as
	the kronecker product of D and E, where D is the temporal covariance
	and E is the spatial covariance.  HQ can then be computed by
	splitting up H into smaller blocks and multiplying those by D columns,
	then by E, and summing up all the blocks.

	This routine is called via the python multiprocessing module, so that
	multiple processes can each work on a section of H, defined by start and end.

	The HQ blocks are stored for later use, and the HQHT for each process is stored.
	The final HQHT can be computed by adding up each of the individual HQHT arrays.

	Input:
		config: configuration options
		start:  starting time step number
		end:    ending time step number
	"""

	ntimesteps = config["ntimesteps"]


	if TC is None:
		print("making temporal covariance values...")
		temporal_cl = config["temporal_corr_length"]
		TC = utils.makeTemporalCovariance(temporal_cl, config)
		print("TC shape is ", TC.shape)

	if SC is None:
		print("getting spatial covariance values...")
		heff = numpy.load(config["workdir"] + "/" + config["sp_cov_file"])
		spatial_cl = config["spatial_corr_length"]
		SC = numpy.matrix(numpy.exp(-heff/spatial_cl))

	# initialize hqht to zero
	hqht = numpy.zeros((config["num_obs"], config["num_obs"]))

	try:
		sigma = numpy.load(sigmafile)
	except:
		try:
			landmask = numpy.load(config["workdir"] + "/" + config["landmask_file"])
			sigma = utils.getSpriorNetcdf(sigmafile, "sigma", landmask)
		except:
			print("Can not read sigma file.")
			sys.exit()


	print("starting ", start, "to", end)
	t0 = datetime.datetime.now()
	for i in range(start, end):
		t1 = datetime.datetime.now()

		# we can do the matrix operations in this loop with the sparse matrices
		hq = csr_matrix((config["num_obs"], config["nlandcells"]))

		for j in range(ntimesteps):

			# skip if temporal covariance is 0
			if TC[i, j] == 0: continue

			# h slice files are numbered 1 .. ntimesteps
			# hsigma files alreay have h*sigma applied
			h = utils.read_sparse_h(j, config, hdir=config["workdir"] + "/" + config["hsigdir"])

			hq = hq + h*TC[j, i]

		hq = hq * SC		# this will create a dense hq
		hq = numpy.array(hq) * sigma[i]		# this will create a dense hq


		# The hq strips are needed later in qht_s, so save them here to file.
		filename = config["workdir"] + "/" + config["hqdir"] + "/HQ%04d" % (i+1)
		numpy.save(filename, hq)

		h = utils.read_sparse_h(i, config)
		hqht = hqht + (hq * h.T)

		t2a = datetime.datetime.now()
		print("Step %4d" % i, "of %4d - %4d" % (start, end), "done. Step time ", t2a-t1, "Total elapsed time ", t2a-t0)


	# each process stores it's own hqht, we'll combine them later.
	name = multiprocessing.current_process().name
	filename = config["workdir"] + "/" + config["hqdir"] + "/HQHT_%s" % (name)
	numpy.save(filename, hqht)

	return


####################################################################

if __name__ == '__main__':

	parser = optparse.OptionParser(usage='%prog [options] receptor_file sigma_file', description="Calculate HQ. ")
	parser.add_option('-m', '--multiprocs', type="int", default=0, help="Use multiple threads.")
	parser.add_option('-c', '--config', default="config.ini", help="set configuration file to use.  Default is 'config.ini'")

	options, args = parser.parse_args()


	if len(args) != 2:
		parser.print_usage()
		sys.exit()


	listfile = args[0]
	sigmafile = args[1]

	try:
		f = open(listfile)
		files = f.readlines()
		f.close()
	except IOError as e:
		sys.exit("File %s cannot open. %s" % (listfile, e))

	numobs = len(files)

	tstart = datetime.datetime.now()

	numpy.set_printoptions(precision=6, edgeitems=10, linewidth=150)

	config = utils.getConfig(options.config)
	config["num_obs"] = numobs

	if not os.path.exists(config["hqdir"]):
		try:
			os.makedirs(config["hqdir"])
		except:
			sys.exit("Can't make directory for hqsub files: %s" % config["hqdir"])

	ntimesteps = config["ntimesteps"]

	if options.multiprocs == 0:
		makeHQBlocks(config, 0, ntimesteps, sigmafile)

		filename = config["workdir"] + "/" + config["hqdir"] + "/HQHT_%s.npy" % ("MainProcess")
		newfilename = config["workdir"] + "/" + config["hqdir"] + "/HQHT.npy"
		os.rename(filename, newfilename)

	else:
		# start nprocs processes for making the hq pieces.
		nprocs = options.multiprocs
		jobs = []
		for i in range(nprocs):
			start = ntimesteps/nprocs*i
			end = ntimesteps/nprocs*(i+1)
			if i == nprocs-1 and end < ntimesteps: end = ntimesteps
			p = multiprocessing.Process(target=makeHQBlocks, args=(config, start, end, sigmafile,))
			jobs.append(p)
			p.start()

		# wait for all the jobs to finish before continuing
		for p in jobs:
			p.join()

		print("combining hqht...")
		# initialize hqht to zero
		hqht = numpy.zeros((config["num_obs"], config["num_obs"]))
		for p in jobs:
			name = p.name
			filename = config["workdir"] + "/" + config["hqdir"] + "/HQHT_%s.npy" % (name)
			a = numpy.load(filename)
			hqht = hqht + a

		filename = config["workdir"] + "/" + config["hqdir"] + "/HQHT"
		numpy.save(filename, hqht)


	tdone = datetime.datetime.now()
	print("Finished. Elapsed time", tdone-tstart)
