#!/usr/bin/python

"""
bayesian inversion

 Solve the equation
  shat = sprior + HQT*inverse(HQHT+R)*(z-hsp)

Input files:
	HQHT.npy, HQHT_pbl.npy, HQHT_trans.npy, HQHT_ftrop.npy
	r.txt
	zhsp.txt
	sprior.npy
	All HQ slice files

Output files:
	hqhti.npy
	hqht_zhsp.npy
	qhts.npy
	hshat.txt
	shat_flux.npy, shat_flux.nc
	shat_pbl.npy, shat_pbl.nc
	shat_trans.npy, shat_trans.nc
	shat_ftrop.npy, shat_ftrop.nc

"""

from __future__ import print_function

import sys
import optparse
import datetime
import numpy

import utils


####################################################################
def qht_s(config, hqht_zhsp, include_boundary, num_boundary_cells=0):
	""" compute hqt*[inverse(hqht+r)*(z-h*sp)]

	inverse(hqht+r)*(z-h*sp) is given as the array hqht_zhsp.
	hq is read in from files created previously.
	"""

	ntimesteps = config["ntimesteps"]

	#--------------
	if include_boundary:
		# the 4200 is the number of cells in the bounday optimization grid * 3 levels (pbl, trans, ftrop)
		qhts = numpy.zeros((ntimesteps, config["locations"] + num_boundary_cells*3))
	else:
		qhts = numpy.zeros((ntimesteps, config["locations"]))
	#--------------

	t0 = datetime.datetime.now()
	for i in range(ntimesteps):
		t1 = datetime.datetime.now()

		# hq slice files are numbered 1 .. ntimesteps
		# hq shape is #obs x #cells
		hqfile = config["workdir"] + "/" + config["hqdir"] + "/HQ%04d.npy" % (i+1)
		hq = numpy.load(hqfile)

		#--------------
		if include_boundary:
			# convert 0-735 3 hour timestep to 1-92 boundary daily timestep number.
			bn = i/8 + 1
			hqpblfile = "HQpbl/HQ%04d.npy" % bn
			hqtransfile = "HQtrans/HQ%04d.npy" % bn
			hqftropfile = "HQftrop/HQ%04d.npy" % bn
			hqpbl = numpy.load(hqpblfile)
			hqtrans = numpy.load(hqtransfile)
			hqftrop = numpy.load(hqftropfile)
			hq = numpy.hstack((hq, hqpbl, hqtrans, hqftrop))
		#--------------

		qhts[i] = numpy.dot(hq.T, hqht_zhsp)

		t2a = datetime.datetime.now()
		print("Step %4d" % (i+1), "of ", ntimesteps, "done. Step time ", t2a-t1, "Total elapsed time ", t2a-t0)

	return qhts


####################################################################
if __name__ == '__main__':

	numpy.set_printoptions(precision=6, edgeitems=10, linewidth=150)

	parser = optparse.OptionParser(usage='%prog [options] receptor_file', description="Do bayesian inversion. ")
	parser.add_option('-b', '--bound', action="store_true", default=False, help="If set, use boundary value optimization files.")
	parser.add_option('-c', '--config', default="config.ini", help="set configuration file to use.  Default is 'config.ini'")
	parser.add_option('-r', '--resume', action="store_true", default=False, help="resume calculations at the qhts step")

	options, args = parser.parse_args()

	if len(args) != 2:
		parser.print_usage()
		sys.exit()

	spriorfile = args[0]
	rfile = args[1]

	config = utils.getConfig(options.config)

	tstart = datetime.datetime.now()


	# Get hqht calculated previously.
	# shape of hqht is numobs x numobs
	print("Getting hqht...")
	filename = config["workdir"] + "/" + config["hqdir"] + "/HQHT.npy"
	hqflux = numpy.load(filename)
	numobs = hqflux.shape[0]
	config["num_obs"] = numobs

	####################
	if options.bound:
		# combine the 4 different hqht's (flux, pbl, trans, free troposphere)
		config_pbl = utils.getConfig("config_pbl.ini")
		config_trans = utils.getConfig("config_trans.ini")
		config_ftrop = utils.getConfig("config_ftrop.ini")

		num_boundary_cells = config_pbl["ncells"]

		filename = config_pbl["workdir"] + "/" + config_pbl["hqdir"] + "/HQHT.npy"
		hqpbl = numpy.load(filename)

		filename = config_trans["workdir"] + "/" + config_trans["hqdir"] + "/HQHT.npy"
		hqtrans = numpy.load(filename)

		filename = config_ftrop["workdir"] + "/" + config_ftrop["hqdir"] + "/HQHT.npy"
		hqftrop = numpy.load(filename)

		hqht = hqflux + hqpbl + hqtrans + hqftrop
	else:
		hqht = hqflux
		num_boundary_cells = 0
	####################


	print("Adding r to hqht...")
	r = numpy.loadtxt(rfile)
	for i in range(numobs):
		hqht[i, i] = hqht[i, i] + r[i]


	# calculate inverse of hqht
	# hqhti is needed later in aposterior variance calculations
	if options.resume:
		print("Loading hqhti...")
		hqhti = numpy.load("hqhti.npy")
	else:
		print("Calculate inverse of hqht...")
		hqhti = numpy.linalg.inv(hqht)
		numpy.save("hqhti", hqhti)


	print("Getting zhsp...")
	zfile = config["workdir"] + "/" + config["z_file"]
	zhsp = numpy.loadtxt(zfile)

	if options.resume:
		print("Loading hqht_zhs...")
		hqht_zhsp = numpy.load("hqht_zhsp.npy")
	else:
		print("Calculating hqhti*zhsp...")
		hqht_zhsp = numpy.dot(hqhti, zhsp)
		numpy.save("hqht_zhsp", hqht_zhsp)

	# compute qht*[inverse(hqht+r)*(z-h*sp)]
	print("Calculating qhts...")
	qhts = qht_s(config, hqht_zhsp, options.bound, num_boundary_cells)
	print("qhts shape is ", qhts.shape)
	numpy.save("qhts", qhts)

	print("Getting sprior...")
	try:
		sp = numpy.load(spriorfile)
	except:
		try:
			landmask = numpy.load(config["landmask_file"])
			sp = utils.getSpriorNetcdf(spriorfile, "sprior", landmask)
		except:
			print("Can not read sprior file.")
			sys.exit()


	if options.bound:
		a = numpy.zeros((config["ntimesteps"], num_boundary_cells*3))
		sp = numpy.hstack((sp, a))
	print("sp shape is ", sp.shape)


	print("Calculate shat...")
	shat = sp + qhts

	print("Saving shat in netcdf format...")
	if options.bound:
		numpy.save("shat_boundary", shat)

		shat_flux = shat[:, 0:config["nlandcells"]]
		shat_pbl = shat[:, config["nlandcells"]:config["nlandcells"] + num_boundary_cells]
		shat_trans = shat[:, config["nlandcells"] + num_boundary_cells:config["nlandcells"] + 2*num_boundary_cells]
		shat_ftrop = shat[:, config["nlandcells"] + 2*num_boundary_cells:config["nlandcells"] + 3*num_boundary_cells]
		numpy.save("shat_flux", shat_flux)
		numpy.save("shat_pbl", shat_pbl)
		numpy.save("shat_trans", shat_trans)
		numpy.save("shat_ftrop", shat_ftrop)
		utils.make_ncdf("shat_flux.nc", "shat", shat_flux, config)
		utils.make_ncdf("shat_pbl.nc", "shat", shat_pbl, config_pbl)
		utils.make_ncdf("shat_trans.nc", "shat", shat_trans, config_trans)
		utils.make_ncdf("shat_ftrop.nc", "shat", shat_ftrop, config_ftrop)

	else:
		numpy.save("shat", shat)
		utils.make_ncdf("shat.nc", "shat", shat, config)


	print("Making hshat...")
	if options.bound:
		hshat = utils.convolve(shat_flux, config)
		hshat_pbl = utils.convolve(shat_pbl, config_pbl)
		hshat_ftrop = utils.convolve(shat_ftrop, config_ftrop)
		hshat_trans = utils.convolve(shat_trans, config_trans)
		numpy.savetxt("hshat_pbl.txt", hshat_pbl)
		numpy.savetxt("hshat_trans.txt", hshat_trans)
		numpy.savetxt("hshat_ftrop.txt", hshat_ftrop)
		numpy.savetxt("hshat_flux.txt", hshat)
	else:
		hshat = utils.convolve(shat, config)
		numpy.savetxt("hshat.txt", hshat)


	tdone = datetime.datetime.now()
	print("Finished. Elapsed time", tdone-tstart)
