#!/usr/bin/python

"""
Script to create the zhsp.txt file.

Takes the input files sprior.npy, obs.txt, bg.txt, calculates
hsprior by convolving sprior with h, then subtracts hsprior and
background values from bg.txt from the observation values in obs.txt
"""

from __future__ import print_function

import sys
import optparse
import numpy

import utils

#########################################################################

epilog = """\
obsfile        : text file with observation value, one line per receptor.
spriorfile     : file with sprior data. Either .npy or netcdf format.
backgroundfile : text file with background value, one line per receptor.
"""

parser = optparse.OptionParser(usage='%prog [options] obsfile spriorfile backgroundfile', description="Calculate zhsp file.", epilog=epilog)
parser.add_option('-o', '--other', default=None, help="text file with other values to subtract from observations, one line per receptor.")
parser.add_option('-c', '--config', default="config.ini", help="set configuration file to use.  Default is 'config.ini'")

options, args = parser.parse_args()

if len(args) != 3:
	parser.print_usage()
	sys.exit()


config = utils.getConfig(options.config)

obsfile = args[0]
spriorfile = args[1]
backgroundfile = args[2]

# Read in observation values.
try:
	obs = numpy.loadtxt(obsfile)
except:
	sys.exit("Error reading observation file: %s" % obsfile)

#print obs
numobs = obs.shape[0]
if obs.ndim > 1: obs = obs[:, 0]


config["num_obs"] = numobs


# -- hsprior
# try reading file as a numpy .npy file first,
# if that doesn't work, try reading it as netcdf file.
print("making hsprior...")
try:
	sprior = numpy.load(spriorfile)
except:
	try:
		landmask = numpy.load(config["landmask_file"])
		sprior = utils.getSpriorNetcdf(spriorfile, "sprior", landmask)
		numpy.save("sprior", sprior)  # save it as a numpy file
	except:
		print("Can not read sprior file.")
		sys.exit()


hsp = utils.convolve(sprior, config)

hspriorfile = config["workdir"] + "/hsprior.txt"
print("saving hsprior in ", hspriorfile)
numpy.savetxt(hspriorfile, hsp)

if options.other:
	hsother = numpy.loadtxt(options.other)


# Get background values previously made using make_bg.py
bg = numpy.loadtxt(backgroundfile)
if bg.ndim > 1: bg = bg[:, 0]


# subtract background values from observation
obsbg = obs - bg
bgfile = "obs-bg.txt"
print("Saving obs - bg in %s..." % bgfile)
numpy.savetxt(bgfile, obsbg, fmt='%12.3f')


# Subtract hother and hsp from bg
print("making zhsp")
if options.other:
	zhsp = obsbg - hsother - hsp
else:
	zhsp = obsbg - hsp


zfile = config["workdir"] + "/zhsp.txt"
print("Saving zhsp in %s..." % zfile)
numpy.savetxt(zfile, zhsp, fmt="%12.6f")
